#! /usr/bin/env python
"""
***************************************************************************
    filename  : Job.py
    author    : Alessandro Maraschini
    email     : egee@datamat.it
    copyright : (C) 2003 by DATAMAT
***************************************************************************
//
// $Id: Job.py,v 1.17.2.16 2007/07/10 12:53:05 acava Exp $
//
"""
import os
import time
from threading import Thread
#import Thread
import socket
import UIutils
from UIutils import exit
from UIutils import errMsg
from glite_wmsui_AdWrapper import AdWrapper
from glite_wmsui_AdWrapper import DagWrapper
from glite_wmsui_LbWrapper import Eve
from glite_wmsui_LbWrapper import Status
import random # for  TCP port range

"""
####################################
                         JobState Class
####################################
"""
class JobState:
  STATE_ID = "StateId"
  CURRENT_STEP="CurrentStep"
  USER_DATA="UserData"
  JOB_STEPS="JobSteps"
  GLITE_JOBID = "edg_jobid"
  state = ""
  """
  Initialise the JobState
  """
  def  __init__(self , state = ""):
    	self.state = AdWrapper( 1 )
  """
  Check the syntax of the State
  """
  def check (self):
	errors = ""
	s_id     = self.state.getStringValue (self.STATE_ID)
	# STATE_ID check
	if len(s_id) >1:
		errors+= self.STATE_ID +": List not allowed\n"
	elif len( s_id) ==1:
		jobid = JobId  ( s_id[0] )
		if not jobid.set:
			errors+= self.STATE_ID +": Wrong Id value\n"
	#  CURRENT_STEP check
	s_tep   = self.state.getStringValue (self.CURRENT_STEP)
	#  USER_DATA check
	u_data = self.state.getStringValue (self.USER_DATA)
	#  JOB_STEPS check
	j_steps = self.state.getStringValue (self.JOB_STEPS)
	if errors:
		UIutils.errMsg("Error","UI_JDL_ERROR", errors)
		exit(1)
  """
  Load a chkpt file
  """
  def fromFile( self ,stateFile ):
	if not os.path.isfile(   stateFile ):
			UIutils.errMsg("Error","UI_FILE_NOT_FOUND",submissionFile)
			exit(1)
	elif ( self.state.fromFile(stateFile) ):
			exit(1)
  """
  Generate a JobState from a string
  """
  def fromString(self, stateStr):
	#self.state.get_error() ;
	self.state.fromString( stateStr.strip() )
	err , apiMsg = self.state.get_error() ;
	if err:
		UIutils.errMsg("Error","UI_JDL_ERROR", "Unable to convert into a jobstate instance.\n"+ apiMsg )
		exit(1)
  """
  toString: convert the State instace into string representation
  """
  def toString( self, multi = 0):
	str = self.state.toLines()
	if not multi:
		str = str.replace ("\n" , " " )
	return str
  """
  Insert the dg_jobid (remove the previous if needed)
  """
  def setId( self, jobid):
	if self.state.hasKey(self.STATE_ID ):
		self.state.removeAttribute( self.STATE_ID )
	self.state.setAttributeStr ( self.STATE_ID , jobid )
  """
  Initialise default JobState from a ready-to-submit jobad
  """
  def initialise ( self, jobad ):
	if  jobad.hasKey ( self.GLITE_JOBID ):
			self.state.setAttributeStr (self.STATE_ID, jobad.getStringValue(self.GLITE_JOBID)[0])
	if  jobad.hasKey ( self.CURRENT_STEP ):
			self.state.setAttributeInt (self.CURRENT_STEP, jobad.getIntValue(self.CURRENT_STEP)[0])
	else:
		self.state.setAttributeInt(self.CURRENT_STEP, 1)
	self.state.setAttributeExpr ( self.USER_DATA , "[]" )
	#### JOB Steps special attribute managing
	if  jobad.hasKey (self.JOB_STEPS ):
			val = jobad.getStringValue (self.JOB_STEPS )
			if val:
				for va in val:
					self.state.addAttributeStr ( self.JOB_STEPS ,  va )
			else: #it's not a string. could be an int
				val =jobad.getIntValue (self.JOB_STEPS )
				if val:
					self.state.setAttributeInt ( self.JOB_STEPS ,  val[0] )
				else: #It's neither a string or an integer but it's present in the jdl: error
					UIutils.errMsg("Error","UI_JDL_ERROR", "Wrong type caught for " + self.JOB_STEPS  )

"""
####################################
                         JobId Class
####################################
"""
class JobId:
  lbHost = ""
  lbPort = 0
  unique = ""
  jobid = ""
  set = 0
  def  __init__(self, jobid=""):
     if jobid:
        self.fromString (jobid)
  #def  __init__(self):
  #   self.set= 0
  """
  create a jobId instance from string
  """
  def  fromString (self, jobid):
    LB_DEFAULT_PROTOCOL = "https://"
    avail_prots=["http","https"]
    output=[]
    if jobid:
     #extract information from jobid, using ':#' as token
     try: #cut http:# from the LB
        prot , body = jobid.split('://',1)
        if not prot in avail_prots:
                return 1
        prot= prot+"://"
        #separate address and  unique string
        lbAddr , unique = body.split('/',1)
        self.lbHost , self.lbPort = lbAddr.split (":" , 1 )
        self.lbPort = int (self.lbPort)
        if  not unique.strip():
               return 1
        self.unique=unique.strip()
        self.set = 1
        self.jobid = jobid

        return 0
     except:
        # display the error message
        return 3
     return 1
"""
  prompt the jobis to the user and let him select among them
  """
def selectJobId (jobid_list , multi = 1):
     if len(jobid_list) ==1:
        return jobid_list
     indJ = 0
     print "------------------------------------------------------------------"
     for jobus in jobid_list:
       indJ = indJ +1
       print repr(indJ).ljust(2) + ':', jobus.jobid
     if multi:
        print 'a'.ljust(2) + ':','all'
     print 'q'.ljust(2) + ':','quit'
     keep=1
     print "------------------------------------------------------------------"
     if multi:
        question= '\nChoose one or more jobId(s) in the list - [1-'+str(indJ)+']all:'
     else:
        question= '\nChoose one jobId in the list - [1-'+str(indJ)+']:'
     jobChoose = []
     while keep:
       ans=raw_input(question)
       ansSplit=ans.split(',')
       anslen = len(ansSplit)
       if (  ans.find(",")* ans.find("-") !=1 ) and not multi:
          keep = 1
       elif (ans==""):
         jobChoose = jobid_list
         keep =0
       elif (ans=="q"):
        print 'bye'
        UIutils.exit(0)
       elif (ans == "a"):
         if multi:
           keep = 0
           jobChoose = jobid_list
       else:
         if (anslen == 1):
           trat = ans.split('-')
           tratlen = len(trat)
           if (tratlen == 1):
             try:
              ind = int(ans)
              if ind in range(1,indJ+1):
               jobChoose.append(jobid_list[ind-1])
               keep=0
              else:
               jobChoose = []
               keep=1
             except:
              jobChoose = []
              keep = 1
           elif (tratlen == 2):
            try:
             s=int(trat[0])
             e=int(trat[1])
             ran = range(s,e+1)
             for k in ran:
              if k in range(1,indJ+1):
                 jobChoose.append(jobid_list[k-1])
                 keep = 0
              else:
                 jobChoose = []
                 keep = 1
            except:
               jobChoose = []
               keep = 1
           else:
             jobChoose = []
             keep = 1
         else:
           for h in ansSplit:
             trat = h.split('-')
             tratlen = len(trat)
             if (tratlen == 1):
              try:
               ind =int(h)
               if ind in range(1,indJ+1):
                jobChoose.append(jobid_list[ind-1])
                keep = 0
               else:
                jobChoose = []
                keep = 1
              except:
               jobChoose = []
               keep = 1
             elif (tratlen == 2):
              try:
               s=int(trat[0])
               e=int(trat[1])
               ran = range(s,e+1)
               for k in ran:
                if k in range(1,indJ+1):
                 jobChoose.append(jobid_list[k-1])
                 keep = 0
                else:
                 jobChoose = []
                 keep = 1
              except:
               jobChoose = []
               keep = 1
             else:
              jobChoose = []
              keep = 1
     if jobChoose:
       print "\n"
       return jobChoose
     else:
        #Unable to find any jobid in the input file
        errMsg(info.logFile,'Error','UI_BAD_INPUT_FILE')
        UIutils.exit(1)
"""
Read the specified input file and return a list containing the jobid (s)
"""
def getJobIdfromFile (inFile ):
     sys_exit = 0
     if not os.path.isfile(inFile):
        errMsg('Error','UI_FILE_NOT_FOUND',inFile)
        exit(1)
     else:
       try:
           f=open(inFile)
       except:
           errMsg('Error','UI_OPEN_ERR',inFile)
           exit(1)
       right_jobs =[]
       str_jobs =[]
       parsed = 0
       for job in f.readlines():
           parsed +=1
           job = job.strip ()
           if (  job.find ('#') != 0) and ( job.find('//') != 0) and (job!=""):
              jobid = JobId()
              if jobid.fromString(job):
                 UIutils.errMsg( 'Warning','UI_WRONG_JOBID_FORMAT',job)
                 sys_exit = -1
              else:
                 if job in str_jobs:
                   UIutils.errMsg( 'Warning','UI_JOBID_REPEAT',job)
                   sys_exit = -1
                 else:
                    right_jobs.append( jobid )
                    str_jobs .append (job)
       f.close
     if not parsed:
          UIutils.errMsg('Error','UI_NO_JOBID')
          exit(1)
     return [sys_exit , right_jobs ]
"""
   Retrieve JobId from a list of strings:
"""
def getJobIdfromList ( jobids):
     jobs= []
     for job in jobids:
       job= job.strip ()
       jobid = JobId ( job )
       if not jobid.set:
          errMsg('Warning','UI_WRONG_JOBID_FORMAT',job)
          pass
       else:
          if job in jobs:
             errMsg('Warning','UI_JOBID_REPEAT',job)
          else:
            jobs.append(jobid)
     return jobs

"""
###############################
  Shadow Class: used to implement listener activities
###############################
"""
class Shadow:
  pipeRoot = ""
  jobid = JobId()
  threadNoGui=""
  #listener = ""
  host = ""
  port = 0
  pid = 0
  refresh  = 1
  accum = ""
  pipe = ""
  pipeOut = ""
  pipeErr = ""
  pipeIn = ""
  barbatruc=""
  writing = 0
  # gui = 1  #Deprecated
  """
  # INitialise parameters
  """
  def set(self , jid):
    self.jobid = jid
    #self.listener = ls
  """
  Write the string into the inputStream
  """
  def write (self, txt):
    self.writing = 1
    if not self.barbatruc:
         self.barbatruc=open ( self.pipeRoot +".in" , "w")
    self.pipeIn = open ( self.pipeRoot +".in" , "w")
    if txt[-1]!="\n":
       txt+="\n"
    self.pipeIn.write (txt)
    self.pipeIn.close()
    self.writing  = 0

  """
  Read and empty the Error buffer
  """
  def emptyErr(self ):
     return  self.pipeErr.read( 1 )
  """
  Read and empty the Output buffer
  """
  def emptyOut(self ):
     if not self.pipeOut:
         self.pipeOut =  open ( self.pipeRoot +".out" )
     return  self.pipeOut.read( 1 )
  """
  get the name of the error pipe
  """
  def getPipeErr(self):
     return self.pipeRoot +".err"
  """
  get the name of the input pipe
  """
  def getPipeIn(self):
     return self.pipeRoot +".in"
  """
  get the name of the  output pipe
  """
  def getPipeOut(self):
      return self.pipeRoot +".out"
  """
  get the name of the local host
  """
  def getHost(self):
     return  self.host
  """
  get the value of the shadow pid
  """
  def getPid(self):
     return self.pid
  """
  get the value of the port which is listening to
  """
  def getPort(self):
      return self.port
  """
  Kill the shadow listener process
  """
  def kill (self):
	if self.pid != 0 :
		command = "kill -9 " + repr( self.pid )
		os.system(command);
  """
  Destroy the instance
  """
  def detach(self):
     if self.barbatruc:
          self.barbatruc.close()
     self.kill()
     for fi in [  self.getPipeErr()  ,  self.getPipeOut()  ,  self.getPipeIn()   ]:
        try:
           os.remove( fi  )
        except:
           pass

  def __call__ (self, command):
	try:
		os.system( command  )
	except KeyboardInterrupt :
		print "bye bye!!!"
		exit(0)

  """
  Check the destination path
  """
  def  checkRootPath(self):
	# Pipe Stream required
	rootDir  = UIutils.FILESEP + "tmp"
	rd = UIutils.solvePath( UIutils.info.confAd.getStringValue ("ListenerStorage")[0] )
	if rd:
		rootDir = rd
	if  not os.path.isdir( rootDir ):
		errMsg('Error','UI_DIR_NOT_FOUND', rootDir)
		exit(1)
	self.pipeRoot =  rootDir +UIutils.FILESEP + "listener-" + self.jobid.unique
	foundPrevious =0
	for fi in [ self.pipeRoot ,  self.getPipeErr() ,   self.getPipeOut() ,  self.getPipeIn() ]:
		if  os.path.exists( fi ):
			errMsg( 'Warning' , 'UI_FILE_EXISTS', fi )
			foundPrevious=1
	if foundPrevious:
		question = "A previous interactive session might still be active.\nDo you whish to proceed anyway?"
		answ=UIutils.questionYN(question)
		if not answ: #No answered
			print "bye"
			exit (0)


  """
  Launch the glite--wms-grid-console-shadow and retrieve host &port information
  """
  def console( self, port=0):
	shPath = UIutils.info.prefix + UIutils.FILESEP +"bin/"
	command =shPath + "glite-wms-grid-console-shadow"
	if  not os.path.isfile( command):
		errMsg('Error','UI_FILE_NOT_FOUND', command)
		exit(1)
	else:
	   command = os.path.abspath( command )
	# Kill the glite-wms-grid-console-shadow, if still running:
	self.kill ()
	arguments = ""

	#Check for opened ports interval due to a possible firewall
	fromPort = 0
	toPort= 0
	env=0
	try:
		env = os.environ["GLOBUS_TCP_PORT_RANGE"]
		#environment variable found
		for sep in [ " " , ":" , "-" , "," ]:
			ind = env.find(sep)
			if ind!=-1:
				try:
					fromPort , toPort = env.split (sep)
					fromPort = int(fromPort)
					toPort = int(toPort) +1
					break
				except:
					pass
		if not fromPort:
			errMsg('Warning',"UI_ENVIRONM_ERROR","GLOBUS_TCP_PORT_RANGE" ,"Unable to parse the value" )
	except:
		pass
	#This variable stores the number of port retrieval.
	#If GLOBUS_TCP_PORT_RANGE is used then UI automatically retry on all available ports before surrending
	retryPort = 1
	#This is the grid_console_shadow error code for an used port
	ALREADY_USED_PORT = 98
	if port:
		# If the port is forced then check the value immediately
		if fromPort and toPort:
			# Port range found at the end of the cycle
			if port<fromPort or port>toPort:
				errMsg('Error','JOB_ATTACH_PORT_RANGE', self.jobid.jobid,  port , env)
				exit(1)
		arguments += " -port " + repr (port)
	elif fromPort and toPort:
		#Port not specified but range found: select randomly a port between the allowed range
		RandomPort = fromPort + int(( toPort - fromPort)* random.random() )
		port = RandomPort
		arguments += " -port " + repr (port)
		retryPort = toPort - fromPort
	# Launch the glite-wms-grid-console-shadow with --logfile (background process)
	# log std in/out/error into named pipe

	for i in range (retryPort):
		arguments += " -log-to-file " + self.pipeRoot   +" &";
		os.system(command + " " + arguments )
		time.sleep(10)
		# get the pid and the port:
		errConsoleInfo = self.getConsoleInfo()
		#Check For Error and if necessary retry
		if errConsoleInfo == ALREADY_USED_PORT:
			if fromPort and toPort and UIutils.info.debug:
				errMsg('Warning', "UI_CAN_NOT_EXECUTE" , shPath +"glite-wms-grid-console-shadow", "Port " + repr(port) +" is already in use")
			if fromPort:
				port =  fromPort +   (port +1) % ( toPort - fromPort )
			arguments = " -port " + repr (port)
		elif errConsoleInfo:
			errMsg('Error', "UI_CAN_NOT_EXECUTE" , shPath +"glite-wms-grid-console-shadow","Command failed")
			exit(1)
		else:
			#No error found. Console successfully Started
			if fromPort and toPort and UIutils.info.debug:
				UIutils.print_message(UIutils.info.logFile,"Firewall range port found. listening port forced to: " +repr (port )  )
			break

	if errConsoleInfo:
			errMsg('Error', "UI_CAN_NOT_EXECUTE" , shPath +"glite-wms-grid-console-shadow","Unable to find any available port")
			exit(1)

  def join(self):
	while 1:
		try:
			time.sleep(2)
		except KeyboardInterrupt :
			print "bye"
			self.detach ()
			exit(0)
	exit(0)

  """
  Once the consloe listener has started, port and pid information
  have to be retrieved from the info named pipe
  """
  def getConsoleInfo(self):
	adStr = "";
	timeout= 0
	MAXTIMEOUT=3
	while 1:
	   try:
	      self.pipe=open (self.pipeRoot)
	      break
	   except:
	      timeout=timeout+1
	      time.sleep(1)
	      if timeout >MAXTIMEOUT :
		errMsg('Error', "UI_CAN_NOT_EXECUTE" , "glite-wms-grid-console-shadow","Timeout limit exceeded while reading info named pipe")
		exit(1)
	while 1:
	    adStr += self.pipe.read( 1 )
	    if  adStr.find("]") !=-1:
	        break ;

	os.remove(   self.pipeRoot )
	ad = AdWrapper ( 1 )
	adStr = adStr.strip()
	if ad.fromString(adStr.replace ("\n" , " " ) ):
	   errMsg('Error', "UI_CAN_NOT_EXECUTE" , "glite-wms-grid-console-shadow","Wrong info pipe stream found")
	   exit(1)
	if ad.hasKey("SHADOW_ERROR"):
		return ad.getIntValue("SHADOW_ERROR")[0];
	port = ad.getIntValue("PORT");
	# Set the Shadow private members
	## PORT
	if len(port) ==1:
	   self.port = port[0]
	else:
	   errMsg('Error', "UI_CAN_NOT_EXECUTE" , "glite-wms-grid-console-shadow", "Wrong retrieved port value")
	   exit(1)
	## PID
	pid  = ad.getIntValue("PID");
	if len(pid) ==1:
	   self.pid = pid[0]
	else:
	   errMsg('Error', "UI_CAN_NOT_EXECUTE" ,  "glite-wms-grid-console-shadow","Wrong retrieved pid value")
	   exit(1)
	## HOST
	self.host = socket.gethostbyaddr(socket.gethostname())[2][0]
	return 0



def __queryEvents__(jid,EVENT_CODE,minLog):
	"""
	Retrieve all required logged event reasons, cutting possible null (nil) reasons
	minLog is the lowest number of events taken into consideration
	i.e. for DONE events at least 2 events are needed (one is already displayed)
	"""
	reasons=[]
	jobid = JobId(jid)
	wrapEvent = Eve( )
	jobEvent = Event(wrapEvent)
	wrapEvent.queryEvents(jobid.lbHost,jobid.lbPort, [jid], \
	#uTAGS, ex ,includes

	[], [], [], [Event.events_codes.index(EVENT_CODE)], "",0,0,0)
	err , apiMsg = wrapEvent.get_error ()
	if err:
		# Do nothing, it is a cosmetic query
		# and may return no element
		return ""
	# Parse reasons
	if wrapEvent.size()>=minLog:
		for ev in range (wrapEvent.size()):
			reasonStr=wrapEvent.getVal(jobEvent.events_names.index("Reason"),ev)[1].strip()
			if reasonStr!="(nil)":
				reasons.append(reasonStr)
	return reasons

"""
###############################
   JobStatus Class
###############################
"""
class JobStatus:

	"""
	Initialise the LB States Names variables
	"""
	states_names=[]
	ATTR_MAX = 0

	"""
	Initialise the LB States Codes variables
	"""
	states_codes=[]
	CODE_MAX = 0

	DSTime = " "  +time.strftime("%Z")

	def  __init__(self, wrStatus=None ):

		self.sons=0
		self.jobidMap={}
		self.jdl=""
		self.level=0
		self.jobid =""
		self.nodeName=""
		
		if not wrStatus is None:
			self.setStatesNames ( wrStatus.getStatesNames() )
			self.setStatesCodes ( wrStatus.getStatesCodes() )

	def __printDoneEvents__(self,jobid,indent,minLog=2):
		"""
		Collect all logged event reasons and create a reasonable string
		jobid is the id of the job to be queried
		indent is the level of indentation for the job
		minLog is the minimal number of logged events to be considered
		(i.e. for DONE jobs 1 reason is already displayed, at least 2 reasons are needed)
		"""
		result=""
		if self.level:
			# Retrieve DONE events
			reasons=__queryEvents__(jobid,"Done", minLog)
			if len(reasons)>=minLog:
				result ="\n" +indent+"Logged Reason(s):"
				for reason in reasons:
					result+="\n".ljust(5)+ indent + "- "+ reason
		return result

	def setStatesNames(self, wrStatesNames):
		# Initialize States Names and append the Tail
		if not self.states_names:
			for st in wrStatesNames:
				st=st.capitalize().replace("_"," ")
				self.states_names.append(st)
			self.states_names.append("Status")
			self.states_names.append("Status Code")
			self.states_names.append("Hierarchy")
			JobStatus.ATTR_MAX = len( self.states_names )

	def setStatesCodes(self, wrStatesCodes):
		# Initialize States Codes and append the Tail
		if not self.states_codes:
			for st in wrStatesCodes:
				st=st.capitalize().replace("_"," ")
				self.states_codes.append(st)
			JobStatus.CODE_MAX = len( self.states_codes )

	def printStatus(self,wrStatus, st , level, noNodes=0):
		info = wrStatus.loadStatus ( st )
		# Map has to be initialised for each PrintStatus
		self.jobidMap={}
		self.level = level
		self.hierarchy = 0
		self.jobid =""
		intervals = int ( len(info) / self.ATTR_MAX  )
		result =""
		for off in range ( intervals ):
			offset = off*self.ATTR_MAX
			result +=self.__printSt__(  info[ offset : offset + self.ATTR_MAX ] )
			if noNodes:
				break
		return result
	def __createJobIdMap__(self):
		if self.jdl and self.jdl.strip().startswith("["):
				dagad = DagWrapper()
				if dagad.fromString(self.jdl):
					err , apiMsg = dagad.get_error ()
					errMsg('Warning','UI_JDL_WRONG_SYNTAX' ,  apiMsg)
				else:
					vectMap=dagad.getMap()
					err , apiMsg = dagad.get_error ()
					if err:
						errMsg('Warning','UI_JDL_WRONG_SYNTAX' ,  apiMsg)
					else:
						for i in range(len(vectMap)/2):
							self.jobidMap[vectMap[i*2]] = vectMap[i*2+1]

	def __printSt__( self, info ):
	
		hierarchy = int(info[self.states_names.index("Hierarchy") ])
		indent = "    "*hierarchy
		stStr = ""
		adStr= ""
		sep = "*************************************************************\n" + indent
		jobid =info[ self.states_names.index("Jobid")]
		if not indent:
			# Indentation is NOT present: main Job
			self.jobid =jobid
			Title= "\n" + indent+""+sep +  "BOOKKEEPING INFORMATION:\n"
			# initialize jdl: this operation is not always present
			self.jdl=info [ self.states_names.index("Jdl") ]
		elif hierarchy > self.hierarchy:
			# Hierarchy increased: previous job was father
			Title = "\n"+ "    "*(hierarchy-1)+"- Nodes information for: "
			if self.level>1:
				Title+= self.jobid+":\n"
			# Managing jobid-node mapping (when possible)
			if not self.jobidMap  and self.jdl:
				# jdl initialised, map still empty
				self.__createJobIdMap__()
		else:
			# Some Indentation Found: it's (not the first) children
			Title= ""
			self.jobid =jobid
		Title = Title + "\n" +indent+"Status info for the Job : " + jobid
		if self.jobidMap:
			if self.jobidMap.has_key(jobid):
				self.nodeName=self.jobidMap[jobid]
			else:
				errMsg('Warning','UI_SUBJOB_ERROR' , "Unable to find node id:"+jobid+"\ninside jdl of dag:\n"+self.jobid)

		s = info [ self.states_names.index("Status") ]
		Title = Title + "\n" + indent+"Current Status:".ljust(20) + s
		#DONE_CODE check (only on Done )
		done_code_str = ""
		if s in ["Aborted","Cancelled"]:
			# minLog is 1, i.e. even one event is taken into consideration
			done_code_str += self.__printDoneEvents__(jobid,indent,1)
		elif s=="Done":
			done_code  = info[ self.states_names.index("Done code") ]
			exit_code    = info[ self.states_names.index("Exit code") ]
			#PARSING Done code and Exit Code
			try:
				done_code  = int ( done_code )
			except ValueError:
				done_code_str = "(Unable to retrieve  a valid done_code)"
			try:
				exit_code  = int ( exit_code )
			except ValueError:
				exit_code = 1
			if done_code== 2:
				done_code_str = "(Cancelled)"
				done_code_str += self.__printDoneEvents__(jobid,indent)
			elif done_code== 1:
				done_code_str = "(Failed)"
				done_code_str += self.__printDoneEvents__(jobid,indent)
			elif done_code== 0 :
				if exit_code==0:
					done_code_str = "(Success)"
					done_code_str += self.__printDoneEvents__(jobid,indent)
				else:
					done_code_str = "(Exit Code !=0)"
			else:
				done_code_str = "(Unexpected done code: " + repr(done_code) +" )"
				done_code_str += self.__printDoneEvents__(jobid,indent)
			if self.level:
				#Add exit code information
				done_code_str = done_code_str +"\n" + indent+"Exit code:".ljust(20) +  repr (exit_code)
		Title = Title + " " + done_code_str
		if self.level:
			reason =  info[ self.states_names.index("Reason") ]
			if reason.strip():
				Title = Title +"\n" + indent+"Status Reason:".ljust(20) + reason
			dest =  info[ self.states_names.index("Destination")]
			if dest.strip():
				Title = Title +"\n" + indent+"Destination:".ljust(20) +  dest

			submissionTime =  int   (info[ self.states_names.index("Stateentertimes")].split(" ")[1].split("=")[1]   )

			Title = Title + "\n" + indent+"Submitted:".ljust(20) + time.asctime(time.localtime( submissionTime )  ) + self.DSTime
			if self.level>2 or hierarchy==0:
				parent = info[ self.states_names.index("Parent job")]
				if parent.strip():
					Title+= "\n" +indent + "Parent Job:".ljust(20) + parent
			#Separator between jdl attributes:
			colsep = "\n"+indent+"      "
		if self.level>1:
		    stStr = stStr + "\n" + indent+"---\n\n" + indent+""
		    for val in range (len (self.states_names)-3 ): #cutting last tree parameters: status, status code, hierarchy
			name = self.states_names[val]
			value = info[val].strip()
			if name and value and value!="(nil)":
				if name in ["Jobid", "Exit code", "Reason", "Parent job"]:
					pass
				elif name == "Condor jdl":
					if self.level>2:
						value = colsep + value.replace ("\n" , colsep )
						adStr += "- "+ name.ljust(24) + "=    " + value +"\n" + indent+""
				elif name in ["Children hist"]:
					if self.level>2:
						value = value.replace( " " , "\n".ljust(6) + indent )
						adStr += "- "+ name.ljust(16) + "=   " + value[0] + "\n".ljust(6) + indent + value[1:] +"\n" + indent+""
				elif name in ["Stateentertime"]:
					stateEnterTime = int (info[ self.states_names.index("Stateentertime")])
					stStr += "- "+ name.ljust(27) + "=    " + time.asctime(time.localtime( stateEnterTime )  ) + self.DSTime +"\n" + indent+""
				elif name == "Stateentertimes":
					if self.level>2:
						values = value.split(" ")
						value = ""
						for val in values[1:-1]:
							nam, val = val.split("=")
							nam = nam.ljust(16)
							try:
								val = int (val)
							except:
								val = 0
							if val==0:
								val = "\t\t---"
							else:
								val = time.asctime(time.localtime( val  ) )+ self.DSTime
							value = value + colsep + nam + " : " + val
						adStr += "- " +name.ljust(16) + "=   " + value +"\n" + indent+""
				elif name in ["Jdl" ]:
					if value[0]=="[":
						#it's a classad
						jobad = AdWrapper (1)
						if jobad.fromString(value):
							err , apiMsg = jobad.get_error ()
							errMsg('Warning','UI_JDL_WRONG_SYNTAX' ,  apiMsg) ;
						else:
							value = jobad.toLines()
							value = colsep + value.replace ("\n" , colsep )
							# Add string only if classad found (and proper level)
							if self.level>2:
								adStr += "- " +name.ljust(16) + "=   " + value +"\n" + indent+""
							if jobad.hasKey("NodeName"):
								self.nodeName=jobad.getStringValue("NodeName")[0]
				elif name in ["Matched jdl"]:
					if self.level>2:
						if value[0]=="[":
							#it's a classad
							jobad = AdWrapper (1)
							if jobad.fromString(value):
								err , apiMsg = jobad.get_error ()
								errMsg('Warning','UI_JDL_WRONG_SYNTAX' ,  apiMsg) ;
							value = jobad.toLines()
							value = colsep + value.replace ("\n" , colsep )
							# Add string only if classad found
							adStr += "- " +name.ljust(16) + "=   " + value +"\n" + indent+""
				elif name in [ "Lastupdatetime"]:
					value = time.asctime(time.localtime(  int (value ) ) )+ self.DSTime
					stStr += "- "+ name.ljust(27) + "=    " + value +"\n" + indent+""
				elif name in ["Rsl"]:
					if self.level>2:
						value = colsep + value.replace ("\n" , colsep )
						adStr += "- " +name.ljust(16) + "=   " + value +"\n" + indent+""
				else:
					stStr += "- "+ name.ljust(27) + "=    " + value +"\n" + indent+""
		if self.nodeName:
			Title = Title + "\n" +indent+"Node Name: ".ljust(20) + self.nodeName
		if adStr:
			adStr = "---\n" + indent+"" +adStr
		stStr= Title + stStr +adStr
		self.hierarchy = hierarchy
		if self.level:
			stStr += "\n" +sep
		return stStr

"""
Event
"""
class Event:

	"""
	Initialise the LB Events Names variables
	"""
	events_names=[]
	ATTR_MAX = 0

	"""
	Initialise the LB Events Codes variables
	"""
	events_codes=[]
	CODE_MAX = 0

	# Separator between jdl attributes:
	sep ="**********************************************************************\n"
	colsep = "\n      "
	evesep = "\n".ljust(14)  +  "***\n"
	DSTime = " " + time.strftime("%Z")
	level  = 0

	def __init__(self, wrEvents = None):
		# DEFAULT constructor
		self.jobidDict={}
		self.lbWrevents = []

		if not wrEvents is None:		
			self.setEventsNames ( wrEvents.getEventsNames() )
			self.setEventsCodes ( wrEvents.getEventsCodes() )

	"""
	PRINT EVENTS methods
	"""
	def printEventsJobId(self, jobid):
		message=""
		# Insert TITLE every Joid
		message += self.sep
		message += "LOGGING INFORMATION:\n\n"
		message += "Printing info for the Job : " + jobid
		if not self.level:
			message+= "\nLogged Events:"
		message += "\n \n"
		timeStamps=self.jobidDict[jobid].keys()
		timeStamps.sort()
		for timeStamp in timeStamps:
			#Ordered timeStamps
			for event in self.jobidDict[jobid][timeStamp]:
				message+=event
		return message

	def setEventsNames(self, wrEventsNames):
		# Initialize Events Names and append the Tail
		if not self.events_names:
			for ev in wrEventsNames:
				ev=ev.capitalize().replace("_"," ")
				self.events_names.append(ev)
			Event.ATTR_MAX = len( self.events_names )

	def setEventsCodes(self, wrEventsCodes):
		# Initialize Events Codes and append the Tail
		if not self.events_codes:
			for ev in wrEventsCodes:
				ev=ev.capitalize().replace("_"," ")
				self.events_codes.append(ev)
			Event.CODE_MAX = len( self.events_codes )

	def printEventsJobIds(self, jobids):
		"""
		Retrieve the stored Logging Information with the provided jobid order
		"""
		message=""
		errorMessage="Unable to retrieve any matching information for jobid: "
		em = "Warning"
		if len (jobids)==1:
			em="Error"
		for jobid in jobids:
			try:
				message +=self.printEventsJobId(jobid.jobid)
			except KeyError, err :
				errMsg ( em , "API_NATIVE_ERROR" , "Eve::queryEvents",errorMessage+jobid.jobid)
		return message
	def printEvents(self):
		"""
		Retrieve the stored Logging Information without a particular Order
		"""
		return self.printDictJobIds(self.jobidDict.keys())
	"""
	UPDATE EVENTS methods
	"""
	def __updateDict__(self,message,ev,timeStamp):
		"""
		Read JobId related to passed ev value and
		if necessary update jobid and timeStamp
		with provided message
		"""
		# TODO each time there is a index access, try/except block is needed
		jobid = self.lbWrevents.getVal ( self.events_names.index("Jobid") ,ev)[1]

		if not self.jobidDict.has_key(jobid):
			# create new Jobid-key instance
			self.jobidDict[jobid]={}
		if not self.jobidDict[jobid].has_key(timeStamp):
			# create new timeStamp-key instance
			self.jobidDict[jobid][timeStamp]=[]
		# Append the event to the list
		self.jobidDict[jobid][timeStamp].append(message)

	def __updateEvent__(self,val,ev):
		"""
		retrieve info for a specific event and attribute
		"""
		stStr   = ""  # Standard messages
		adStr   = ""  # Ad messages
		timeStamp=0
		name , value = self.lbWrevents.getVal (val,ev)
		# capitalized and underscore replaced:
		name  = name.strip().capitalize().replace("_"," ")
		value = value.strip()
		if name and value and value!="(nil)":
			if name in ["Jobid", "Event"]:
				pass
			elif name in [ "Timestamp"]:
				timeStamp = int(value)
				value = time.asctime(time.localtime(  int (value ) ) )
				stStr += "- "+ name.ljust(27) + "=    " + value +self.DSTime+"\n"

			elif name in [ "Arrived"]:
				if  self.level>1:
					value = time.asctime(time.localtime(  int (value ) ) )
					stStr += "- "+ name.ljust(27) + "=    " + value+self.DSTime+"\n"
			elif name in ["Job" , "Jdl", "Classad"]:
				if (  self.level >2):
					if value[0]=="[":
						jobad = AdWrapper (1)
						if jobad.fromString(value):
							err , apiMsg = jobad.get_error()
							errMsg('Warning','UI_JDL_WRONG_SYNTAX' , apiMsg ) ;
						value = jobad.toLines()
						value = self.colsep + value.replace( "\n" , self.colsep )
					adStr += "- " +name.ljust(15) + "=   " + value +"\n"
			elif name in ["Level", "Seqcode"]:
				if (  self.level >2):
					stStr += "- "+ name.ljust(27) + "=    " + value +"\n"
			elif name in ["Priority"]:
				if (  self.level >2):
					try:
						if int(value):
							value="synchronous"
						else:
							value="asynchronous"
					except:
						pass
					stStr += "- "+ name.ljust(27) + "=    " + value +"\n"
			elif name in ["Source","Destination","Result", "Dest id"]:
				stStr += "- "+ name.ljust(27) + "=    " + value +"\n"
			elif  self.level>1:
				stStr += "- "+ name.ljust(27) + "=    " + value +"\n"
		return [stStr, adStr,timeStamp]

	def updateEvents(self,wrapEvent,  level):
		"""
		retrieve info for all the events
		This method will call __updateEvent__
		and then will call __updateDict__
		"""
		self.lbWrevents = wrapEvent
		self.level  = level
		# Cycling Events:
		for ev in range ( wrapEvent.size() ):
			message   = ""  # Single event Message
			stStr     = ""  # Standard messages
			adStr     = ""  # Ad messages
			timeStamp = 0   # Event TimeStamp
			if self.level:
				message += "\t---\n"
				message += "Event: "
			else:
				message +=" - "
			# message += self.lbWrevents.getVal(self.events_names.index("event"),ev)[1]+"\n"
			message += self.lbWrevents.getEventName(ev)[1]+"\n"
			if self.level:
				# Cycling event attributes inside a single event (only when level>0):
				for val in range (len (self.events_names)):
					prOut = self.__updateEvent__(val,ev)
					# collect normal and AD messages into two different pools
					stStr += prOut[0]
					adStr += prOut[1]
					if prOut[2]:
						# timeStampFound
						timeStamp=prOut[2]
				# join the two pools
				message += stStr+adStr
			# Update JobId Dictonary with all the collected events for this instance of wrapEvent
			self.__updateDict__(message,ev,timeStamp)
